#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/capability.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/fs.h>
/* Just to get the AIO/EPOLL method */
#include <linux/aio.h> 
#include <linux/file.h>
#include <linux/ioctl.h>
#include <linux/nfs_fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

static int exastore_major = 0;

#define EXA_MISC_VERSION    "1.0"
#define EXA_MISC_DESC       "ExaStore additions to kernel"
#define NAME "exaioctl"


struct nfs_getfh {
  	int fd;
	struct nfs_fh *fh;
};

#define EXANET_NFSIOCTL         _IOR('e', 1000, struct nfs_getfh)
#define EXANET_PFMEMALLOCSET    _IOWR('e', 1001, unsigned long)
	  
static int exastore_ioctl(struct inode *inode,
					  	  struct file *file,
					  	  unsigned int cmd,
					  	  unsigned long arg)
{
	switch (cmd) {

	case EXANET_NFSIOCTL: {
		int rc = -EBADF;
		struct file *filp;
		struct inode *inode;
		struct nfs_getfh getfh;
		struct files_struct *files;
		struct nfs_getfh __user *user_getfh = (struct nfs_getfh __user *) arg;
		
		if (copy_from_user(&getfh, (void *) arg, sizeof(getfh)))
			return -EFAULT;

#ifdef DEBUG
		printk(KERN_DEBUG "received NFSIOCTLCMD from task %d on fd = %d\n", current->pid, getfh.fd);
#endif

		filp = fget(getfh.fd);
		if (filp) {
			inode = filp->f_dentry->d_inode;
			if ( inode->i_sb->s_magic == NFS_SUPER_MAGIC) {
				getfh.fh = NFS_FH(inode);
				if (getfh.fh) {
					if(copy_to_user((void*) user_getfh->fh, getfh.fh,
							getfh.fh->size + ((int) &((struct nfs_fh*)0)->data)) == 0) {
						rc = 0;
					} else {
						rc = -EFAULT;
					}	
				} 
			}
			fput(filp);
		}
		return rc;
	}
	case EXANET_PFMEMALLOCSET: {
		int enable;

		if (copy_from_user(&enable, (void *) arg, sizeof(int)))
			return -EFAULT;

		if (enable) 
			current->flags |= (PF_MEMALLOC | PF_MEMALLOC_USER);
		else
			current->flags &= ~(PF_MEMALLOC | PF_MEMALLOC_USER);

		printk(KERN_DEBUG "[%d] Setting PF_MEMALLOC to %d\n", current->pid, enable);

		break;
	}
	default:
		return -EINVAL;
	}

	return 0;
}

#ifdef CONFIG_COMPAT
#include <linux/compat.h>

struct compat_nfs_getfh {
        compat_int_t fd;
        compat_caddr_t fh;
};

#define EXANET_NFSIOCTL32         _IOR('e', 1000, struct compat_nfs_getfh)
#define EXANET_PFMEMALLOCSET32    _IOWR('e', 1001, unsigned)

long compat_exastore_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
        long  err;
	
	switch (cmd) {
		case EXANET_NFSIOCTL32:  {
			struct compat_nfs_getfh *compat_getfh;
			struct nfs_getfh *getfh;
			compat_uptr_t fh;
			int32_t fd;

			compat_getfh = (struct compat_nfs_getfh *) arg;
			err = get_user(fd, &compat_getfh->fd);
			err |= get_user(fh, &compat_getfh->fh);
			if (!err) {
				getfh = compat_alloc_user_space(sizeof(struct nfs_getfh));
				if (getfh) {
					put_user(fd, &getfh->fd);
					put_user(compat_ptr(fh), &getfh->fh);

					err = exastore_ioctl(NULL, NULL, EXANET_NFSIOCTL, (unsigned long) getfh);
				} else {
					err = -ENOMEM;
				}
			} else {
				err = -EFAULT;	
			}
			break;
		}
		case EXANET_PFMEMALLOCSET32:
			err = exastore_ioctl(NULL, NULL, EXANET_PFMEMALLOCSET, arg);
			break;
		default:
			printk("compat_exastore_ioctl: invalid argument (%x)\n", cmd);
			err = -EINVAL;
	}

	return err;
}
#endif

static struct file_operations exastore_fops =
{
	.owner      = THIS_MODULE,
	.open       = nonseekable_open,
	.ioctl      = exastore_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = compat_exastore_ioctl,
#endif
};

static int __init exa_ioctl_init(void)
{
	int r;

	r = register_chrdev(exastore_major, NAME, &exastore_fops);
	if (r < 0) {
		printk(KERN_ERR NAME ": unable to register character device\n");
		return r;
	}
	if (!exastore_major) {
		exastore_major = r;
		printk(KERN_DEBUG NAME ": got dynamic major %d\n", exastore_major);
	}

	return 0;
}

static void __exit exa_ioctl_cleanup(void)
{
	unregister_chrdev(exastore_major, NAME);
}

module_init(exa_ioctl_init);
module_exit(exa_ioctl_cleanup);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(EXA_MISC_DESC ", v" EXA_MISC_VERSION);
MODULE_AUTHOR("ExaNet Ltd., <support@exanet.com>");

