/* This is a module which is used to remote sysrq. */
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/capability.h>
#include <linux/init.h>
#include <linux/sysrq.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <net/checksum.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/workqueue.h>
#include <linux/ctype.h>

#include "rsysrq.h"
#include "rsysrq_handle.h"

#define SYSRQ_WAIT_TIMEOUT (HZ * 5)

static int rsysrqd_pid = -1;
static DECLARE_COMPLETION(rsysrqd_exited);
static DECLARE_WAIT_QUEUE_HEAD(rsysrqd_wqh);

static sysrq_delayed_t sysrq_delayed_list;
static rwlock_t sysrq_delayed_lock = RW_LOCK_UNLOCKED;

static void rsysrq_fire(void)
{
	struct list_head *pos, *q;
	sysrq_delayed_t sysrq_execute;
	sysrq_delayed_t *delayed_code = NULL;

	/*
	 * Nothing to do 
	 */
	if (list_empty(&sysrq_delayed_list.list))
		return;

	/* 
	 * SysRq may take a long time (especially sysrq-t), so copy to temporary list 
	 * and release the lock ASAP.
	 */
	INIT_LIST_HEAD(&sysrq_execute.list);
	write_lock(&sysrq_delayed_lock);
	list_for_each_safe(pos, q, &(sysrq_delayed_list.list)) {
		delayed_code = list_entry(pos, sysrq_delayed_t, list);
		if (delayed_code) {
			list_del(pos);
			list_add_tail(&(delayed_code->list), &(sysrq_execute.list));
		}
	}
	write_unlock(&sysrq_delayed_lock);	

	/* 
	 * Now do the actual magic keys work 
	 */
	list_for_each_safe(pos, q, &(sysrq_execute.list)) {
		delayed_code = list_entry(pos, sysrq_delayed_t, list);
		if (delayed_code) {
			list_del(pos);
			handle_sysrq(delayed_code->c, NULL, NULL);
			kfree(delayed_code);
		}
	}
}

static
int target_do(char *buf, int len)
{
	unsigned char c;
	struct timeval now;
	int err = 0;
	sysrq_delayed_t *new_code = NULL;
	unsigned long flags;

	if (len != sizeof(struct rsysrq_s)) 
		return 0;

	do_gettimeofday(&now);
	if ((c = rsysrq_extract((struct rsysrq_s*)buf, passwd, now.tv_sec, tolerance)) == 0)
		return 0;

	write_lock_irqsave(&sysrq_delayed_lock, flags);

	/*
	 * 'B', 'C' Reboot requests are done immediately.
	 * 'b', 'c' Reboot requests are entered to the queue like all other requests. 
	 */
	if (c == 'B' || c == 'C')  {
		write_unlock_irqrestore(&sysrq_delayed_lock, flags);
		handle_sysrq(tolower(c), NULL, NULL);
	} else {
		int add_code = 1;
		sysrq_delayed_t *delayed_code = NULL;
		struct list_head *pos, *q;
		err = -ENOMEM;

                /* Do we already have this key in the delayed list */
        	list_for_each_safe(pos, q, &(sysrq_delayed_list.list)) {
                	delayed_code = list_entry(pos, sysrq_delayed_t, list);
                	if (delayed_code) {
				if (c == delayed_code->c) {
					err = 0;
					add_code = 0;
					break;
				}
			}
                }

		if (add_code) {	
			new_code = kmalloc(sizeof(sysrq_delayed_t), GFP_ATOMIC);
			if (new_code) {
				err = 0;
				new_code->c = c;
				list_add_tail(&(new_code->list), &(sysrq_delayed_list.list));
				write_unlock_irqrestore(&sysrq_delayed_lock, flags);
			} else {
				write_unlock_irqrestore(&sysrq_delayed_lock, flags);
			}
		} else {
			write_unlock_irqrestore(&sysrq_delayed_lock, flags);
		}
	}

	return (err >= 0);
}

unsigned int
target(struct sk_buff **pskb,
       const struct net_device *in,
       const struct net_device *out,
       unsigned int hooknum,
       const struct xt_target *target,
       const void *targinfo,
       void *userinfo)
{
        const struct iphdr *iph = (*pskb)->nh.iph;
        if(iph->protocol==IPPROTO_UDP)
        {
                struct udphdr *udp = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
                char *data= (char*)udp + sizeof(struct udphdr);
                int len = ntohs(udp->len) - sizeof(struct udphdr);
                printk(KERN_DEBUG "Remote SysRq: ip=%u.%u.%u.%u sport=%d dport=%d len=%d\n",
                        ((unsigned char*)(&(iph->saddr)))[0],
                        ((unsigned char*)(&(iph->saddr)))[1],
                        ((unsigned char*)(&(iph->saddr)))[2],
                        ((unsigned char*)(&(iph->saddr)))[3],
                        ntohs(udp->source),ntohs(udp->dest),len);

                if(!target_do(data, len)) {
                        printk(KERN_NOTICE "Remote SysRq: Unauthorized attempt to remote sysrq!!!"
                                " (ip=%u.%u.%u.%u sport=%d dport=%d len=%d)\n",
                                ((unsigned char*)(&(iph->saddr)))[0],
                                ((unsigned char*)(&(iph->saddr)))[1],
                                ((unsigned char*)(&(iph->saddr)))[2],
                                ((unsigned char*)(&(iph->saddr)))[3],
                                ntohs(udp->source),ntohs(udp->dest),len);
                }
        }

        return IPT_RETURN;
}

static int rsysrqd_thread(void *data)
{
	int ret;

	daemonize("rsysrqd");
	allow_signal(SIGKILL);

	do {
		/*
		 * Wait for list to be non empty.
		 */
		ret = wait_event_interruptible_timeout(rsysrqd_wqh, 
							 !list_empty(&sysrq_delayed_list.list), 
							 SYSRQ_WAIT_TIMEOUT);
		if (ret) {
			rsysrq_fire();
		}

	} while (!signal_pending(current));

	complete_and_exit(&rsysrqd_exited, 0);
}

int init_rsysrq_wq(void)
{
	int pid = -1;

	INIT_LIST_HEAD(&sysrq_delayed_list.list);
	pid = kernel_thread(rsysrqd_thread, NULL,
			    CLONE_FS | CLONE_FILES | CLONE_SIGHAND);

	if (pid < 0) {
		printk(KERN_WARNING "Remote SysRq: failed to create rsysrqd_thread\n");
		return -1;
	}

	rsysrqd_pid = pid;

	return 0;
}

void kill_rsysrq(void) 
{
	int ret;

	ret = kill_proc(rsysrqd_pid, SIGKILL, 1);
	wait_for_completion(&rsysrqd_exited);
}

